<h2>Table of Contents</h2>
{% if section.subsections -%}
    <ul>
        {% for subsec in section.subsections -%}
            {% set sec_ = get_section(path=subsec) -%}
            <li>
                <a href="{{ sec_.permalink | safe }}{{ _ugly_url }}">{{ sec_.title }}</a>
                {% if sec_.pages -%}
                    <ul>
                        {% for page_ in sec_.pages -%}
                            <li><a href="{{ page_.permalink | safe }}{{ _ugly_url }}">{{ page_.title }}</a></li>
                        {% endfor %}
                    </ul>
                {% endif -%}
            </li>
        {% endfor %}
    </ul>
{% else -%}
    <h3> No Sections Found </h3>
{% endif -%}
