{% extends "index.html" %}
{% block title %} {{ config.title }} | {{ page.title }} {% endblock title %}
    {% block content %}
        <h1>{{ page.title }}</h1>
        {{ page.content | safe }}
    {% endblock content %}
