{% extends "index.html" %}

{% block title %}{{ config.title }} | 404 Page Not Found{% endblock title %}

{% block content %}
    <h2 class="title">404: Page Not Found</h2>
    <p>The page you requested could not be found.</p>
    <p><a href="{{ config.base_url }}">Return to Home</a></p>
{% endblock content %}
