# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import os as _os

pkgdatadir = _os.path.join(_os.path.dirname(_os.path.dirname(__file__)), "share", "meshy")
QR_SCANNER_ENABLED = True
SHORTCUTS_DIALOG_ENABLED = True
