# Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors
# SPDX-License-Identifier: GPL-3.0-or-later

import sys

import gi

gi.require_version("Gtk", "4.0")
gi.require_version("Adw", "1")

from meshy.application import MeshyApplication


def main(version):
    app = MeshyApplication(version=version)
    return app.run(sys.argv)
