.search-container {
  display: none;

  &--is-visible {
    display: block;
    width: 100%;
  }

  #search {
    width: 100%;
    display: block;
    border:none;
    border-left: 1px solid $color;
    padding:1px 0;
    text-align: left;
    line-height: $baseline;
    font-size: $font-size;
    font-family:$font-family;
    color:$color;
    background:transparent;
  }

  #search:focus {
    outline:none;
    border:none;
  }

  .search-results {
    &__header {
      font-weight: bold;
      padding: 1rem 0rem;
    }

    &__items {
      margin: 0 2vw;
      padding: 0;
      list-style: circle;
    }

    &__item {
      margin-bottom: 1rem;
    }

    &__teaser {

    }
  }
}

#on_right {
  display: block;
  text-align: right;
  margin-bottom: $baseline;
}

#search-ico {
  font-family: 'FabricMDL2Icons';
  cursor: pointer;
  font-size: $baseline;
  line-height: 1;
}
