/* Copyright 2026, Jiri Eischmann <jiri@eischmann.cz> and the meshy contributors */
/* SPDX-License-Identifier: GPL-3.0-or-later */

.outgoing-message {
    background-color: alpha(@accent_bg_color, 0.15);
    border-color: alpha(@accent_bg_color, 0.3);
}

searchbar > revealer > box {
    background: none;
    box-shadow: none;
    border: none;
    padding-top: 2px;
    padding-bottom: 2px;
}

.navigation-sidebar row {
    margin-left: 4px;
    margin-right: 4px;
    padding-left: 4px;
    padding-right: 4px;
}

.nav-unread-dot {
    background-color: @accent_bg_color;
    border-radius: 50%;
    min-width: 6px;
    min-height: 6px;
    padding: 0;
    font-size: 0;
}

.unread-badge {
    background-color: @accent_bg_color;
    color: @accent_fg_color;
    border-radius: 12px;
    min-width: 12px;
    font-size: 0.8em;
    font-weight: bold;
    padding: 2px 5px 1px;
}

.emoji-avatar {
    background-color: alpha(@accent_bg_color, 0.15);
    border-radius: 50%;
    min-width: 32px;
    min-height: 32px;
    font-size: 18px;
}

.unread-divider {
    color: @accent_bg_color;
    font-weight: bold;
    font-size: 0.8em;
}

.date-divider-line {
    min-height: 1px;
    margin-top: 0;
    margin-bottom: 0;
}

.unread-divider-line {
    min-height: 1px;
    margin-top: 0;
    margin-bottom: 0;
    background-color: @accent_bg_color;
}

.map-marker-icon {
    color: white;
}

.map-marker-badge {
    background-color: rgba(60, 60, 60, 0.85);
    color: white;
    font-size: 8px;
    font-weight: bold;
    min-width: 14px;
    min-height: 14px;
    border-radius: 7px;
    padding: 0;
}

.map-overlay-box {
    background: alpha(@window_bg_color, 0.85);
    border-radius: 6px;
}

.map-overlay-label {
    background: alpha(@window_bg_color, 0.85);
    border-radius: 6px;
    padding: 4px 12px;
}

.nav-rail {
    background-color: @sidebar_bg_color;
    border-right: 1px solid alpha(@borders, 0.5);
    padding-top: 4px;
    padding-left: 4px;
    padding-right: 3px;
}

listview.background > row {
    min-height: 0;
    padding: 0;
    border-radius: 0;
    background: none;
    outline: none;
}

.pill {
    border-radius: 16px;
    padding-left: 16px;
    padding-right: 16px;
}

.theme-tile {
    border-radius: 12px;
    min-width: 100px;
    min-height: 50px;
    box-shadow: 0 0 0 1px alpha(@borders, 0.5);
}

button.flat:checked .theme-tile {
    box-shadow: 0 0 0 2px @accent_bg_color;
}
