$baseline: 1.5rem;

$background: #f9fafb;
$color: #1f2937;

$links: #2563eb;
$hover-links: #1d4ed8;
$visited-links: #2563eb;

$font-size: 1.05rem;
$font-family: system-ui, -apple-system, "Segoe UI", Roboto, sans-serif;
$line-height: 1.75;
$code_font: 400 0.95rem/1.6 "JetBrains Mono", "Fira Code", "Cascadia Code", monospace;

$quote-bg: #eff6ff;
$quote-border: #2563eb 4px solid;
$quote-alert-note-color: #2563eb;
$quote-alert-tip-color: #10712b;
$quote-alert-important-color: #7544d0;
$quote-alert-warning-color: #915f00;
$quote-alert-caution-color: #bf1626;
